# Purpose: Setup and initialize the Dock Panes gui interfaces
#
# $Id: dockpane.rb,v 1.8 2004/10/19 22:13:47 ljulliar Exp $
#
# Authors:  Laurent Julliard <laurent AT moldus DOT org>
# Contributors:
#
# This file is part of the FreeRIDE project
#
# This application is free software; you can redistribute it and/or
# modify it under the terms of the Ruby license defined in the
# COPYING file.
#
# Copyright (c) 2002 Laurent Julliard. All rights reserved.
#

module FreeRIDE
  module FoxRenderer
    
    ##
    # This is the module that renders dockpanes using
    # FOX.
    #
    class DockPane
      extend FreeBASE::StandardPlugin
      ICON_PATH = "/system/ui/icons/DockPane"

      def DockPane.start(plugin)
        component_slot = plugin["/system/ui/components/DockPane"]
        
        component_slot.subscribe do |event, slot|
          if (event == :notify_slot_add && slot.parent == component_slot)
            Renderer.new(plugin, slot)
          end
        end
        
        component_slot.each_slot { |slot| slot.notify(:notify_slot_add) }

        # Handle icons
        plugin[ICON_PATH].subscribe do |event, slot|
          if event == :notify_slot_add
            app = plugin['/system/ui/fox/FXApp'].data
            path = "#{plugin.plugin_configuration.full_base_path}/icons/#{slot.name}.png"
            if FileTest.exist?(path)
              slot.data = Fox::FXPNGIcon.new(app, File.open(path, "rb").read)
              slot.data.create
            end
          end
        end
    
        # Now only is the plugin ready
        plugin.transition(FreeBASE::RUNNING)
      end
      
      ##
      # Each instance of this class is responsible for rendering a dockpane component
      #
      class Renderer
        include Fox
        
        FOCUS_COLOR = Fox::FXRGB(0,84,227)
        FOCUS_TEXT_COLOR = FXColor::GhostWhite
        DEFAULT_COLOR = 4292405740
        DEFAULT_TEXT_COLOR = FXColor::Black
        
        attr_reader :plugin
        
        def initialize(plugin, slot)
          @plugin = plugin
          @slot = slot
          
          @main_window = @plugin["/system/ui/fox/FXMainWindow"].data
          @icons = @plugin[ICON_PATH]

          @frame = FXVerticalFrame.new(@main_window, FRAME_NONE|LAYOUT_FILL_Y|LAYOUT_FILL_X,0,0,0,0,0,0,0,0,0,0)

          @sub_frame = FXVerticalFrame.new(@frame, FRAME_NONE|LAYOUT_FILL_Y|LAYOUT_FILL_X,0,0,0,0,0,0,0,0,0,0)

          title_bar = FXHorizontalFrame.new(@sub_frame, FRAME_NONE|LAYOUT_FILL_X,0,0,0,0,0,0,0,0,0,0)
          @title = FXLabel.new(title_bar, @slot.name, nil, JUSTIFY_LEFT|LAYOUT_FILL_Y|LAYOUT_FILL_X|FRAME_NONE,0,0,0,0)

          @title.setFont(FXFont.new(@plugin["/system/ui/fox/FXApp"].data, "helvetica", 8, FONTWEIGHT_NORMAL))
          @title.setTextColor(FOCUS_TEXT_COLOR)
          @title.frameStyle = FRAME_LINE
          @title.setBorderColor(FOCUS_COLOR)
          @title.setBackColor(FOCUS_COLOR)
          @title.connect(SEL_LEFTBUTTONPRESS) {@slot.data.setFocus}

          # dock/undock and close buttons
          btn_style = (FRAME_NONE|LAYOUT_FILL_Y)
          @dock_btn = FXButton.new(title_bar,"\tClose\tClose the view", @icons['undock_dockpane'].data, nil, 0, btn_style,0,0,0,0)
          @dock_btn.setBackColor(FOCUS_COLOR)
          @close_btn = FXButton.new(title_bar,"\tClose\tClose the view", @icons['close_dockpane'].data, nil, 0, btn_style,0,0,0,0)
          @close_btn.setBackColor(FOCUS_COLOR)

          # create the frame
          @frame.create

          # manage events
          @frame.connect(SEL_FOCUSIN) {
            @title.setTextColor(FOCUS_TEXT_COLOR)
            @title.setBackColor(FOCUS_COLOR)
            @title.setBorderColor(FOCUS_COLOR)
            @close_btn.setBackColor(FOCUS_COLOR)
            @dock_btn.setBackColor(FOCUS_COLOR)
          }
          @frame.connect(SEL_FOCUSOUT) {
            @title.setTextColor(DEFAULT_TEXT_COLOR)
            @title.setBackColor(DEFAULT_COLOR)
            @title.setBorderColor(DEFAULT_COLOR)
            @close_btn.setBackColor(DEFAULT_COLOR)
            @dock_btn.setBackColor(DEFAULT_COLOR)
          }

          @close_btn.connect(SEL_COMMAND, method(:onCmdClose))

          @dock_btn.connect(SEL_COMMAND, method(:onCmdDockUndock))

          @slot.subscribe do |event, slot|
            update(event) if ((event == :refresh) || (event == :notify_data_set && slot==@slot))
          end
          
          setup_actions
          
          @plugin.log_info << "DockPane Renderer for #{@slot.path} created"
        end
        
        def setup_actions
          bind_action("dock", :dock)
          bind_action("undock", :undock)
          bind_action("show", :show)
          bind_action("hide", :hide)
          bind_action("current?", :current?)
        end
        
        def bind_action(name, meth)
          @slot["actions/#{name}"].set_proc method(meth)
        end
        
        ### Commands ###
        
        def dock(dockbar_path)

          # for a redock the dockpane must be undocked AND a
          # dialog box must exist otherwise it is first time we dock (see below)
          if !@slot.manager.docked? && !@dlg_box.nil?
            @sub_frame.reparent(@frame)
            @dock_btn.icon = @icons['undock_dockpane'].data
            show
            return
          end

          # First time we dock
          @dockbar_path = dockbar_path
          @tabbook = @slot["/system/ui/fox/dockbar/#{where}/tabbook"].data
          @tab = FXTabItem.new(@tabbook, @slot.name, nil,FRAME_NONE,0,0,0,0,0,0,0,0)
          # keep track of the dockpane manager in the Widget
          @tab.userData = @slot.manager

          if @tabbook.tabStyle==TABBOOK_BOTTOMTABS
            @tab.tabOrientation = TAB_BOTTOM
          else
            @tab.tabOrientation = TAB_TOP
          end
          @tab.hide
          @frame.hide
          @tab.create
          @frame.reparent(@tabbook)
          @tab.connect(SEL_FOCUSIN) {@slot.data.setFocus}
          @plugin.log_info << "Reparenting #{@slot.path} to #{dockbar_path}"
        end
        
        def undock
          # Create a separate dialog box the dockpane will appear
          app = @plugin["/system/ui/fox/FXApp"].data

          # make the dockpane disappear from the dockbar
          hide
          
          # create the dialog box and reparent the dockpane frame to it
          @dlg_box = FXDialogBox.new(app,  @slot.name, DECOR_ALL, 0, 0, @sub_frame.width, @sub_frame.height,0,0,0,0,0,0)
          @dlg_box.connect(SEL_CLOSE, method(:onCmdClose))

          @dlg_box.create
          @sub_frame.reparent(@dlg_box)
          @dock_btn.icon = @icons['dock_dockpane'].data
          @dlg_box.show
          @slot.notify(:refresh)
        end

        def show
          if @slot.manager.docked?
            @tab.show
            @frame.show
            @tabbook.current = @tabbook.indexOfChild(@tab)
            @tab.recalc
            @frame.recalc
            @slot.data.setFocus
          else
            if @dlg_box
              # @dlg_box.raise FIXME: (FXRuby says it's a private method, why????)
              @dlg_box.setFocus
            end
          end
        end
        
        def hide
          @tab.hide
          @frame.hide
          @tab.recalc
          @frame.recalc
        end

        def where
          @slot[@dockbar_path].name
        end

        ##
        # Check whether this is the currently active dockpane
        # 
        def current?
          # current means it is docked, the tab is not hidden
          # and it is the current tab
          @tab && @tab.shown && (@tabbook.current*2 == @tabbook.indexOfChild(@tab))
        end

        ##
        # Called whenever the dockpane may need to be updated
        # in particular when the .data of the slot is
        # modified (meaning that a new graphical object must be
        # inserted in the dockpane
        # 
        def update(event)
          # New visual object to be inserted in the tab frame.
          # slot.data contains the path to the UI component to insert
          # the FOX frame to insert is in the attr_frame attribute
          if event == :notify_data_set and @slot.data
            child_frame = @slot.data
            child_frame.reparent(@sub_frame)
            child_frame.show
          end
        end
        
        def onCmdClose(sender, sel, ptr)
          if @slot.manager.docked?
            @slot.manager.hide
          else
            # if undocked, redock it first and then hide
            onCmdDockUndock(sender, sel, ptr)
            @slot.manager.hide
          end
          return 1
        end

        def onCmdDockUndock(sender, sel, ptr)
          if @slot.manager.docked?
            @slot.manager.undock
          else
            # close the separate dialog box, redock and 
            # show in dockpane again
            @dlg_box.hide
            @slot.manager.dock(where)
            @slot.manager.show
            @dlg_box.destroy
            @dlg_box = nil
          end
          return 1
        end
        
      end
    end
    
  end
end
